/* SYNTAX

program    	->  	PROGRAM declaration* BEGIN statements END   

declaration   	->	PROC ID ( ID ) BEGIN statements END   
		|	ID ;

block  		->	statement
		-> 	( statements )

statements 	->	statement+   

statement 	->	SKIP ;
		|	ID := aexpression ;
		|	IF bexpression THEN block [ELSE block]
		|	WHILE bexpression DO block
		|	CALL ID ( aexpression ) ;   

aexpression 	->	ID
		|	NUM
		|	aexpression + aexpression
		|	aexpression - aexpression
		|	aexpression * aexpression
		|	aexpression / aexpression
		|	- aexpression
		|	( aexpression )   

bexpression 	-> 	TRUE
		|	FALSE
		|	aexpression < aexpression
		|	aexpression <= aexpression
		|	aexpression > aexpression
		|	aexpression >= aexpression
		|	aexpression = aexpression
		|	aexpression <> aexpression
		|	NOT bexpression
		|	bexpression AND bexpression
		|	bexpression OR bexpression
		|	( bexpression )

*/

import java.io.*;

public class Parser 
  {
  private Lexer L;  
  private Token lookahead;
  private Token prev;
  
/* ------------------------------------------------------------------------- */

  public Parser(Lexer l) 
    {
    L = l;
    try
      {
      lookahead = L.next(); 
      prev = lookahead;
      } 
    catch (Exception e)
      {
      e.printStackTrace();
      }
    }

/* ------------------------------------------------------------------------- */

  private SyntaxException syntacticError(String whatWeExpected)
    {
    return new SyntaxException("Expected " + whatWeExpected + " instead of `" + lookahead.spelling + "'; " + lookahead.range + ".");
    }

/* ------------------------------------------------------------------------- */
  
  private void match(int expectedType) throws SyntaxException, IOException
    {
    if (lookahead.type == expectedType) {
      prev = lookahead;
      lookahead = L.next();
    }
    else
      throw syntacticError(Token.spellingOf[expectedType]);
    }

  private void matchIt() throws SyntaxException, IOException
    { prev = lookahead; lookahead = L.next(); }

/* ------------------------------------------------------------------------- */
// Program
/* ------------------------------------------------------------------------- */

  public Program parseProgram() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);

    match(Token.PROGRAM);
    Decls decls = parseDecls();
    match(Token.BEGIN);
    Statements ss = parseStatements();
    match(Token.END);
    match(Token.EOF);
    r.finish(prev.range.last);
    return new Program(r, decls, ss);
    }

/* ------------------------------------------------------------------------- */
// Declarations
/* ------------------------------------------------------------------------- */
  private Decls parseDecls() throws SyntaxException, IOException
    {
    Decls ds = new Decls(lookahead.range);
    while (lookahead.type != Token.BEGIN) 
      if (lookahead.type == Token.PROC) 
        ds.append(parseProcDecl());
      else 
        ds.append(parseVarDecl());
    ds.finish(prev.range.last);
    return ds;
    }

  private VarDecl parseVarDecl() throws IOException, SyntaxException
    {
    Range r = new Range(lookahead.range.first);
    Id  name = parseIdentifier();
    match(Token.SEMICOLON);
    r.finish(prev.range.last);
    return new VarDecl(r, name);
    }

  private ProcDecl parseProcDecl() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);
    match(Token.PROC);
    Id name = parseIdentifier();
    match(Token.LPAREN);
    Range r1 = new Range(lookahead.range.first);
    Id formalname = parseIdentifier();
    r1.finish(prev.range.last);
    FormalParam formal = new FormalParam(r1, formalname);
    match(Token.RPAREN);
    match(Token.BEGIN);
    Statements ss = parseStatements();
    match(Token.END);
    r.finish(prev.range.last);
    return new ProcDecl(r, name, formal, ss);
    }

/* ------------------------------------------------------------------------- */
// Block 
/* ------------------------------------------------------------------------- */

   private Statement parseBlock() throws IOException, SyntaxException
    {
    Statement s;
    if (lookahead.type == Token.LPAREN)
      {
      Range r = new Range(lookahead.range.first);
      matchIt();
      Statements ss = parseStatements();
      match(Token.RPAREN);
      r.finish(prev.range.last);
      s = new Block(r, ss);
      }
    else
      s = parseStatement();
    return s;
    }
     

/* ------------------------------------------------------------------------- */
// Statements
/* ------------------------------------------------------------------------- */

  private Statements parseStatements() throws IOException, SyntaxException
    {
    Statements ss = new Statements(lookahead.range);
    while (lookahead.type != Token.END && lookahead.type != Token.RPAREN)
      ss.append(parseStatement());
    ss.finish(prev.range.last);
    return ss;
    }

/* ------------------------------------------------------------------------- */

  private Statement parseStatement() throws IOException, SyntaxException
    {
    switch (lookahead.type)
      {
      case Token.SKIP      : return parseSkipStmt();
      case Token.ID        : return parseAssignStmt();
      case Token.IF        : return parseIfStmt();
      case Token.WHILE     : return parseWhileStmt();
      case Token.CALL      : return parseCallStmt();
      default              : throw syntacticError("statement");
      }
    }

/* ------------------------------------------------------------------------- */

  private SkipStmt parseSkipStmt() throws IOException, SyntaxException
    {
    Range r = new Range(lookahead.range.first);
    match(Token.SKIP);
    match(Token.SEMICOLON);
    r.finish(prev.range.last);
    return new SkipStmt(r);
    }

/* ------------------------------------------------------------------------- */

  private AssignStmt parseAssignStmt() throws IOException, SyntaxException
    {
    Range r = new Range(lookahead.range.first);
    Id dest = parseIdentifier();
    match(Token.ASSIGN);
    Expression source = parseAdditiveExpr();
    match(Token.SEMICOLON);
    r.finish(prev.range.last);
    return  new AssignStmt(r, dest, source);
    }

/* ------------------------------------------------------------------------- */

  private IfStmt parseIfStmt() throws IOException, SyntaxException
    {
    Range r = new Range(lookahead.range.first);
    match(Token.IF);
    Expression premise = parseExpression();
    match(Token.THEN);
    Statement conclusion = parseBlock();
    Statement alternative = null;
    if (lookahead.type == Token.ELSE)
      {
      matchIt();
      alternative = parseBlock();
      }
    r.finish(prev.range.last);
    return new IfStmt(r, premise, conclusion, alternative);
    }

/* ------------------------------------------------------------------------- */

  private WhileStmt parseWhileStmt() throws IOException, SyntaxException
    {
    Range r = new Range(lookahead.range.first);    
    match(Token.WHILE);
    Expression requirement = parseExpression();
    match(Token.DO);
    Statement body = parseBlock();
    r.finish(prev.range.last);
    return new WhileStmt(r, requirement, body);
    }

/* ------------------------------------------------------------------------- */

  private CallStmt parseCallStmt() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);
    match(Token.CALL);
    Id name = parseIdentifier();
    match(Token.LPAREN);
    Expression ap = parseAdditiveExpr();
    match(Token.RPAREN);
    match(Token.SEMICOLON);
    r.finish(prev.range.last);
    return new CallStmt(r, name, ap);
    }
 
/* ------------------------------------------------------------------------- */
// Expressions.
/* ------------------------------------------------------------------------- */

  private Id parseIdentifier() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);
    String spelling = lookahead.spelling.toString();
    match(Token.ID);
    r.finish(prev.range.last);
    return new Id(r, spelling);
    }

  private Id parseIdentifierExpr() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);
    String spelling = lookahead.spelling.toString();
    match(Token.ID);
    r.finish(prev.range.last);
    return new IdExpr(r, spelling);
    }
 
/* ------------------------------------------------------------------------- */

  private Num parseNumber() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);
    String spelling = lookahead.spelling.toString();
    match(Token.NUM);
    r.finish(prev.range.last);
    return new Num(r, spelling);
    }

/* ------------------------------------------------------------------------- */

  private Bool parseBoolean() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);
    String spelling = lookahead.spelling.toString();
    match(Token.BOOL);
    r.finish(prev.range.last);
    return new Bool(r, spelling);
    }

/* ------------------------------------------------------------------------- */

  private Expression parsePrimaryExpr() throws SyntaxException, IOException
    {
    if (lookahead.type == Token.NUM)
      return parseNumber();
    else if (lookahead.type == Token.ID)
      return parseIdentifierExpr();
    else if (lookahead.type == Token.LPAREN)
      {
      matchIt();
      Expression e = parseAdditiveExpr();
      match(Token.RPAREN);
      return e;
      }
    else
      throw syntacticError("expression");
   }
/* ------------------------------------------------------------------------- */

    private Expression parseUnaryExpr() throws SyntaxException, IOException
      {
      if(lookahead.type == Token.NOT
         || lookahead.type == Token.MINUS)
        {
        Range r= new Range(lookahead.range.first);
        int op = lookahead.type;
        matchIt();
        Expression right = parsePrimaryExpr();
        r.finish(prev.range.last); 
        return new UnaryExpr(r,op,right);
        } 
      else 
        {   
        return parsePrimaryExpr();
        }
      }

/* ------------------------------------------------------------------------- */

  private Expression parseMultiplicativeExpr() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);    
    Expression left = parseUnaryExpr();
    while (lookahead.type == Token.TIMES || lookahead.type == Token.DIV)
      {
      int op = lookahead.type;
      matchIt();
      Expression right = parseUnaryExpr();
      r.finish(prev.range.last);
      left = new ArithmeticExpr(r, op, left, right);
      }
    return left;
    }

/* ------------------------------------------------------------------------- */

  private Expression parseAdditiveExpr() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);    
    Expression left = parseMultiplicativeExpr();
    while (lookahead.type == Token.PLUS || lookahead.type == Token.MINUS)
      {
      int op = lookahead.type;
      matchIt();
      Expression right = parseMultiplicativeExpr();
      r.finish(prev.range.last);
      left = new ArithmeticExpr(r, op, left, right);
      }
    return left;
    }

/* ------------------------------------------------------------------------- */

  private Expression parseRelationalExpr() throws SyntaxException, IOException
    {
    Range r= new Range(lookahead.range.first);
	
    Expression left = parseAdditiveExpr();
    while (lookahead.type == Token.LT || lookahead.type == Token.GT
        || lookahead.type == Token.LTEQ || lookahead.type == Token.GTEQ)
      {
      int op = lookahead.type;
      matchIt();
      Expression right = parseAdditiveExpr();
      r.finish(prev.range.last);
      left = new RelationalExpr(r,op,left,right);
      }
    return left;
    }

/* ------------------------------------------------------------------------- */

  private Expression parseEqualityExpr() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);    
    Expression left = parseRelationalExpr();
    while (lookahead.type == Token.EQ || lookahead.type == Token.NOTEQ)
      {
      int op = lookahead.type;
      matchIt();
      Expression right = parseRelationalExpr();
      r.finish(prev.range.last);
      left = new EqualityExpr(r, op, left, right);
      }
    return left;
    }


/* ------------------------------------------------------------------------- */

  private Expression parsePrimaryBoolExpr() throws SyntaxException, IOException
    {
    if (lookahead.type == Token.BOOL)
      return parseBoolean();
    else if (lookahead.type == Token.LPAREN)
      {
      matchIt();
      Expression e = parseExpression();
      match(Token.RPAREN);
      return e;
      }
    else
      return parseEqualityExpr();
   }

/* ------------------------------------------------------------------------- */

  private Expression parseUnaryBoolExpr() throws SyntaxException, IOException
    {
    if(lookahead.type == Token.NOT)
      {
      Range r= new Range(lookahead.range.first);
      int op = lookahead.type;
      matchIt();
      Expression right = parsePrimaryBoolExpr();
      r.finish(prev.range.last); 
      return new UnaryExpr(r,op,right);
      } 
    else
      {
      return parsePrimaryBoolExpr();
      }
    }

/* ------------------------------------------------------------------------- */

  private Expression parseConditionalAndExpr() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);    
    Expression left = parseUnaryBoolExpr();
    while (lookahead.type == Token.AND)
      {
      matchIt();
      Expression right = parseUnaryBoolExpr();
      r.finish(prev.range.last);
      left = new ConditionalExpr(r, Token.AND, left, right);
      }
    return left;
    }

/* ------------------------------------------------------------------------- */

  private Expression parseExpression() throws SyntaxException, IOException
    {
    Range r = new Range(lookahead.range.first);    
    Expression left = parseConditionalAndExpr();  
    while (lookahead.type == Token.OR)
      {
      matchIt();
      Expression right = parseConditionalAndExpr();
      r.finish(prev.range.last);
      left = new ConditionalExpr(r, Token.OR, left, right);
      }
    return left;
    }

  }
